open Error
open Xlist

open Interfaces

(** class of grammar constructors *)
class constructor_impl _gmod _name _params _reg : constructor =
  object (self)
      (** grammar module of this element *)
    val mutable gmod : gmodule = _gmod
	(** position in the source file *)
    val reg : Error.region = _reg
	
	(** constructor name *)
    val name : string = _name
	(** parameter list *)
    val params : parameter list = _params
	
	(** accessors *)
    method get_name : string = name
    method get_params : parameter list = params
	
	(** post-initialization with the given gmodule *)
    method set_gmodule (gm : gmodule) : unit =
      begin
	(** set the gmod field *)
	gmod <- gm;
	(** checks whether the constructor contains several indexes *)
	(if List.length (List.find_all (fun p -> p#is_index) params) > 1
	then Error.fail reg ("multiple indices not allowed") else ());
	(** performs post_initialization for parameters *)
	List.iter (fun p -> p#set_gmodule gm) params
      end
	
	(** returns true/false whether this is an index constructor *)
    method is_index : bool =
      List.exists (fun p -> p#is_index) params
	
	(** returns the index name *)
    method get_index_name : string =
      (List.find (fun p -> p#is_index) params)#get_name
	    
	    (** returns the through list of the category of each parameter
	     * along with the categories themselves if they have index *)
    method get_throughs : grammar list =
      Xlist.doublon_remove
	(List.concat
	   (List.map (fun p -> p#get_throughs) params))
	
	(** induction pattern for this constructor *)
    method induction_pattern (g : grammar) : string list =
      List.concat (List.map (fun p -> p#induction_pattern g) params)
	
	(** rewrite induction hypothesis for this constructor *)
    method rewrite_induction (prop : property) (g : grammar) : string =
      (* we put first the index parameters in order to call #index_tac# first
	 (calling #crush# first causes le_lt_dec to appear...) *)
      let params_with_index_first =
	Xlist.concat_pair (List.partition (fun p -> p#is_index) params) in
  (* generation of induction proof script for each parameter *)
      let params_ind_pre =
	List.map (fun p -> p#rewrite_induction prop g) params_with_index_first in
  (* removing duplicated calls to the same rewriting, which causes reversion... *)
      let params_ind = Xlist.doublon_remove params_ind_pre in
  (* removing empty element in params_ind to avoid "; ;" *)
      let non_empty_params_ind =
	List.filter (fun s -> s <> "") params_ind in
  (* if not an index parameter, we can always begin with simplifying and calling f_equal *)
  (* TODO "simpl; f_equal" should not work if the property is not an equality ! *)
      (if not self#is_index then "simpl; f_equal; " else "") ^
      String.concat "; " non_empty_params_ind ^
  (* we add #crush# after all parameter scripts *)
      (if non_empty_params_ind = [] then "#crush#" else "")
	
	(** function equation generation *)
    method fun_equation_gen (def : fun_definition) : string list =
      def#constructor_function_equation_gen (self :> constructor)
	
	(** Pretty printing *)
    method to_string () : string =
      "  | " ^ name ^ (String.concat "" (List.map (fun p -> p#to_string ()) params))
  end
