open Xlib
open Xstring
open Xlist

open Interfaces

(** Abstract class for syntax definitions *)
class virtual abstract_terms_definition _gmod : terms_definition =
  object (self)
      (** the grammar module currently processed *)
    val gmod : gmodule = _gmod
	
	(** the target grammar actually processed *)
    val mutable target_grammar : grammar option = None
    method private set_target (g : grammar) : unit = target_grammar <- Some g
    method get_target : grammar = Xlib.some_of target_grammar

	(** definition title *)
    method virtual title : string

	(** extra definition lines to be inserted before the grammar definitions *)
    method coq_predefs : string list = []

	(** extra definition lines to be inserted after the constructor cases *)
    method aux_defs : string list = []

	(** the category of generated index case *)
    method virtual index_type : string
	
	(** category names decoration (to avoid conflict) *)
    method decorate_type_name (name : string) : string = name
	
	(** constructors names decoration (to avoid conflict) *)
    method decorate_constr_name (name : string) : string = name
	
	(** generation of shifts *)
    method virtual shift_terms_gen : (grammar * shift_exp) -> string
	
	(** generation of parameters *)
    method parameter_terms_gen (p : parameter) : string =
      if p#is_index then "(" ^ p#get_name ^ " : " ^ self#index_type ^ ")"
      else
	let shifts =
	  String.concat " "
	    (List.map self#shift_terms_gen p#get_shifts) in
	let shifts_final = if shifts = "" then "" else shifts ^ " " in
	shifts_final ^ "(" ^ p#get_name ^ " : " ^
	self#decorate_type_name p#get_type#get_name ^ ")"

(** generation of constructors *)
    method constructor_terms_gen (c : constructor) : string =
      let subst =
	("#constr#", self#decorate_constr_name c#get_name) ::
	("#params#",
	 String.concat
	   " " (List.map self#parameter_terms_gen c#get_params)) ::
	[]
      and text =
	"  | #constr# #params#"
      in
      Xstring.replace subst text
	
	(** generation of grammars *)
    method grammar_terms_gen (prefix : string) (g : grammar) : string list =
      let _ = self#set_target g in
      let subst =
	("#Inductive#", prefix) ::
	("#raw_type#", g#get_name) ::
	("#type#", self#decorate_type_name g#get_name) ::
	[]
      and substs =
	("#constr_lines#",
	 List.map self#constructor_terms_gen g#get_constrs) ::
	("#aux_lines#", self#aux_defs) ::
	[]
      and text =
	"  #Inductive# #type# : Type :=" ::
	"#constr_lines#" ::
	"#aux_lines#" ::
	[] in
      Xstring.replaces subst substs text
  end

class virtual abstract_fun_definition _gmod : fun_definition =
  object (self)
      (** the grammar module currently processed *)
    val gmod : gmodule = _gmod
	
	(** the source grammar actually processed *)
    val mutable source_grammar : grammar option = None
    method private set_source (g : grammar) : unit = source_grammar <- Some g
    method get_source : grammar = Xlib.some_of source_grammar
	
	(** the target grammar actually processed *)
    val mutable target_grammar : grammar option = None
    method private set_target (g : grammar) : unit = target_grammar <- Some g
    method get_target : grammar = Xlib.some_of target_grammar
	
	(** the definition title *)
    method virtual title : string
	
	(** extra definition lines to be inserted before the grammar definitions *)
    method coq_predefs : string list = []

	(** extra definition lines to be inserted after the constructor cases *)
    method aux_defs : string list = []

	(** function parameters list (name * type_name) *)
    method virtual fun_params : (string * string) list	
	
	(** return type of the function *)
    method virtual return_type : string
	
	(** radix of the function name *)
    method virtual fun_radix : string
	
	(** name of the function in a given target grammar *)
    method fun_name_type_name (name : string) : string =
      (match source_grammar with
	None -> ""
      | Some g -> g#get_name ^ "_") ^ self#fun_radix ^ "_in_" ^ name
								   
								   (** checks the validity of the function *)
    method is_valid : bool =
      (match source_grammar with
	None -> true
      | Some g -> List.mem (Some g) self#get_source_grams) &&
      (match target_grammar with
	None -> false
      | Some g -> List.mem g (self#get_target_grams gmod#get_all_grams))
	
	(** name of the function *)
    method fun_name : string =
      self#fun_name_type_name self#get_target#get_name
	
	(** category names decoration (to avoid conflict) *)
    method decorate_type_name : string -> string = fun x -> x
	
	(** constructors names decoration (to avoid conflict) *)
    method decorate_constr_name : string -> string = fun x -> x
	
	(** the source grammars for which generetion is needed. *)
    method virtual get_source_grams : grammar option list
	
	(** the target (processed) grammars for which generetion is needed.
	 * By default, every grammar with the source grammar in its throughs. *)
    method virtual get_target_grams : grammar list -> grammar list
	
	(** check if the type of the parameter is a self recursion *)
    method is_self_rec (p : parameter) : bool =
      p#get_type = self#get_target
	
	(** check if the type of the parameter is a back recursion *)
    method is_back_rec (p : parameter) : bool =
      p#get_type = self#get_source
	
	(** check if the type of the parameter is a throughs recursion *)
    method is_throughs_rec (p : parameter) : bool =
      List.mem self#get_source (p#get_type#get_throughs)
	
	(** check if the type of the parameter is a module recursion *)
    method is_module_rec (p : parameter) : bool =
      gmod#is_gmodule_gram p#get_type#get_name
	
	(** generation of shifts *)
    method virtual shift_rhs_function_gen : (grammar * shift_exp) -> string
	
	(** generation of parameters *)
    method virtual parameter_rhs_function_gen : parameter -> string

	(** generation of constructors rhs : index case *)
    method virtual index_constructor_rhs_function_gen : constructor -> string list
	
	(** generation of constructors rhs : non index case *)
    method non_index_constructor_rhs_function_gen (c : constructor) : string list =
      [ String.concat " "
	  (c#get_name :: (List.map (fun p -> self#parameter_rhs_function_gen p) c#get_params)) ]
	
	(** generation of constructors rhs *)
    method constructor_rhs_function_gen (c : constructor) : string list =
      if c#is_index
      then self#index_constructor_rhs_function_gen c
      else self#non_index_constructor_rhs_function_gen c

 	  (** generation of shifts *)
    method shift_lhs_function_gen : (grammar * shift_exp) -> string = fun _ -> ""
	
	(** generation of parameters *)
    method typed_parameter_lhs_function_gen (p : parameter) : string * string =
      (p#get_name, self#decorate_type_name p#get_type#get_name)
    method typed_parameters_lhs_function_gen (c : constructor) : (string * string) list =
      List.map self#typed_parameter_lhs_function_gen c#get_params
	
	(** generation of constructors lhs *)
    method constructor_lhs_function_gen (c : constructor) : string =
      (self#decorate_constr_name c#get_name) ^ " " ^
      (String.concat " " (List.map fst (self#typed_parameters_lhs_function_gen c)))

	(** generation of constructors *)
    method constructor_function_gen (c : constructor) : string list =
      let subst =
	("#lhs#", self#constructor_lhs_function_gen c) ::
	[]
      and substs =
	("#rhs#", List.map (fun s -> "        " ^ s) (self#constructor_rhs_function_gen c)) ::
	[]
      and text =
	"      | #lhs# =>" ::
	"#rhs#" ::
	[] in
      Xstring.replaces subst substs text
	
	(** generation of constructors fun equation *)
    method constructor_function_equation_gen (c : constructor) : string list =
      let typed_constr_params =
	List.map (fun (x, t) -> "(" ^ x ^ " : " ^ t ^ ")") (self#typed_parameters_lhs_function_gen c) in
      let typed_fun_params =
	List.map (fun (x, t) -> "(" ^ x ^ " : " ^ t ^ ")") self#fun_params in
      (** builds the forall header with above constructors and function parameters, or
	  nothing if empty (empty forall statement leads to an error in Coq) *)
      let forall_header =
	if typed_constr_params <> [] || typed_fun_params <> []
	then ["    forall " ^ (String.concat " " typed_fun_params) ^
	      " " ^ (String.concat " " typed_constr_params) ^ ","]
	else [] in
      let subst =
	("#lhs#", self#constructor_lhs_function_gen c) ::
	("#fun_params#",
	 String.concat " " (List.map fst self#fun_params)) ::
	("#fun_name#", self#fun_name) ::
	("#lemma_name#", self#fun_name ^ "_" ^ c#get_name) ::
	[]
      and substs =
	("#rhs#", List.map (fun s -> "        " ^ s) (self#constructor_rhs_function_gen c)) ::
	[]
      and text =
	"  Lemma #lemma_name# :" ::
	forall_header @
	"      #fun_name# #fun_params# (#lhs#) =" ::
	"#rhs#" ::
	"  ." ::
	"  Proof." ::
	"    intros; simpl; auto." ::
	"  Qed." ::
	"" ::
	"  Hint Rewrite #lemma_name# : #database#." ::
	[] in
      Xstring.replaces subst substs text
	
	(** generation of function for a given target grammar *)
    method grammar_function_gen (prefix : string) (target : grammar) : string list =
      let _ = self#set_target target in
      let subst =
	("#Fixpoint#", prefix) ::
	("#raw_target_type#", target#get_name) ::
	("#target_type#", self#decorate_type_name target#get_name) ::
	("#source_type#",
	 match source_grammar with
	   Some g -> self#decorate_type_name g#get_name
	 | None -> "") ::
	("#fun_name#", self#fun_name) ::
	("#return_type#", self#return_type) ::
	("#fun_params#",
	 String.concat " "
	   (List.map (fun (x, t) -> "(" ^ x ^ " : " ^ t ^ ")") self#fun_params)) ::
	[]
      and substs =
	("#constr_lines#",
	 List.concat (List.map self#constructor_function_gen target#get_constrs)) ::
	("#aux_lines#", self#aux_defs) ::
	[]
      and text =
	"  #Fixpoint# #fun_name# #fun_params# (_arg : #target_type#) { struct _arg } : #return_type# :=" ::
	"    match _arg with" ::
	"#constr_lines#" ::
	"#aux_lines#" ::
	"    end" ::
	[]
      in
      Xstring.replaces subst substs text

    method index_constructor_lemmas_and_tactics : constructor -> (string list * string list) =
      failwith "Definitions_abstract.index_constructor_lemmas_and_tactics: illegal call"

    method private index_lemmas_and_tactics (g : grammar) : (string list * string list) =
      let _ = self#set_source g in
      let _ = self#set_target g in
      let index_constrs = List.filter (fun c -> c#is_index) g#get_constrs in
      let (lemmas, tactics) =
	List.split (List.map self#index_constructor_lemmas_and_tactics index_constrs) in
      let subst = ("#type#", g#get_name) :: [] in
      (Xstring.replaces subst [] (List.concat lemmas),
       Xstring.replaces subst [] (List.concat tactics))

    method index_constructor_equation_lemmas : constructor -> (string list * string list) =
      failwith "Definitions_abstract.index_constructor_equation_lemmas: illegal call"

    method private index_equation_lemmas (g : grammar) : (string list * string list) =
      let _ = self#set_source g in
      let _ = self#set_target g in
      let index_constrs = List.filter (fun c -> c#is_index) g#get_constrs in
      let (lemmas, tactics) =
	List.split (List.map self#index_constructor_equation_lemmas index_constrs) in
      let subst = ("#type#", g#get_name) :: [] in
      (Xstring.replaces subst [] (List.concat lemmas),
       Xstring.replaces subst [] (List.concat tactics))

    method index_tactic_name : string option = None

    method index_lemmas_and_tactics_gen : string list =
      match self#index_tactic_name with
	None -> []
      | Some tactic_name ->
	  let sources = self#get_source_grams in
	  
	  let process_source (source : grammar option) =
	    let gram = Xlib.some_of source in
	    let _ = self#set_source gram in
	    let targets = gram :: [] in
	    
	    let index_lemmas_and_tacs =
	      List.split
		(List.map self#index_lemmas_and_tactics targets) in
	    
	    let index_equations_tacs =
	      List.split
		(List.map self#index_equation_lemmas targets) in
	    
	    let lemmas = List.concat (fst index_lemmas_and_tacs) in
	    let tac_bodies = List.concat (snd index_lemmas_and_tacs) in
	    let lemmas' = List.concat (fst index_equations_tacs) in
	    let tac_bodies' = List.concat (snd index_equations_tacs) in
	    
	    (lemmas, tac_bodies, lemmas', tac_bodies')
	  in
	  
	  let (lemmas, tac_bodies, lemmas', tac_bodies') =
	    Xlist.split_4 (List.map process_source sources) in

	  let all_lemmas = List.concat lemmas in
	  let all_tac_bodies = List.concat tac_bodies in
	  let all_lemmas' = List.concat lemmas' in
	  let all_tac_bodies' = List.concat tac_bodies' in
	  
	  let subst =
	    ("#index_dec_tac#", tactic_name ^ "_dec") ::
	    ("#index_tac#", tactic_name) ::
	    [] in
	  let substs =
	    ("#index_lemmas#", all_lemmas) ::
	    ("#tactic_body#", all_tac_bodies) ::
	    ("#index_equation_lemmas#", all_lemmas') ::
	    ("#equation_tactic_body#", all_tac_bodies') ::
	    []
	  and text =
	    "#index_lemmas#" ::
	    "" ::
	    (if all_tac_bodies <> [] then
	      "  Ltac #index_dec_tac# :=" ::
	      "    match goal with" ::
	      "#tactic_body#" ::
	      "    end." :: []
	    else []) @
	    "" ::
	    "#index_equation_lemmas#" ::
	    "" ::
	    (if all_tac_bodies' <> [] then
	      "  Ltac #index_tac# :=" ::
	      "    repeat match goal with" ::
	      "#equation_tactic_body#" ::
	      "    end." :: []
	    else []) in
	  Xstring.replaces subst substs text
  end
