open Xstring

open Interfaces
open Definitions_abstract

class de_bruijn_terms _gmod =
  object (self)
    inherit abstract_terms_definition _gmod
	
	(** the definition title *)
    method title : string = "De Bruijn primitive syntax"
	(** the category of generated index case *)
    method index_type : string = "nat"
	
	(** generation of shifts *)
    method shift_terms_gen (type_name, level) : string = ""
  end

class lift_fun _gmod =
  object (self)
    inherit abstract_fun_definition _gmod
	
	(** the definition title *)
    method title : string = "De Bruijn lifting function"
	(** function parameters list (name * type_name) *)
    method fun_params : (string * string) list =
      ("_n", "nat") :: ("_m", "nat") :: []
					 
					 (** return type of the function *)
    method return_type : string =
      self#get_target#get_name
	
 	(** radix of the function name *)
     method fun_radix : string = "lift"
 
     method index_tactic_name : string option = Some "lift_index_tac"

						 (** the source grammars for which generetion is needed.
						  * In this case : every indexed grammar. *)
    method get_source_grams : grammar option list =
      List.map (fun g -> Some g) gmod#get_indexed_grams
	
	(** the target (processed) grammars for which generetion is needed.
	 * By default, every grammar with the source grammar in its throughs. *)
    method get_target_grams (grams : grammar list) : grammar list =
      List.filter
	(fun g ->
	  (g = self#get_source) ||
	  (List.mem self#get_source g#get_throughs))
	grams
	
	(** generation of shifts *)
    method shift_rhs_function_gen (gram, level) : string =
      if gram = self#get_source
      then string_of_shift level
      else ""
	  
	  (** generation of parameters *)
    method parameter_rhs_function_gen (p : parameter) : string =
      let shift =
	let raw_shifts = List.map self#shift_rhs_function_gen p#get_shifts in
	let non_empty_shifts = List.filter (fun s -> s <> "") raw_shifts in
	String.concat " + " non_empty_shifts in
      let subst =
	("#param#", p#get_name) ::
	("#recfun#", self#fun_name_type_name p#get_type#get_name) ::
	("#shift#", if shift = "" then "_m" else "(_m + " ^ shift ^ ")") ::
	[]
      and text =
	if self#is_self_rec p || self#is_back_rec p || self#is_throughs_rec p
	then
	  "(#recfun# _n #shift# #param#)"
	else
	  "#param#" in
      Xstring.replace subst text
	
	(** generation of constructors rhs : index case *)
    method index_constructor_rhs_function_gen (c : constructor) : string list =
      let id_params =
	String.concat " "
	  (List.map (fun p -> self#parameter_rhs_function_gen p) c#get_params) in
      let mod_params =
	String.concat " "
	  (List.map (fun p ->
	    if p#is_index then "(" ^ p#get_name ^ " + _n)"
	    else self#parameter_rhs_function_gen p) c#get_params) in
      let subst =
	("#constr#", c#get_name) ::
	("#index#", c#get_index_name) ::
	("#id_params#", id_params) ::
	("#mod_params#", mod_params) ::
	[]
      and text =
	if self#get_target = self#get_source
	then
	  "if le_lt_dec _m #index# then #constr# #mod_params# else #constr# #id_params#" ::
	  []
	else "#constr# #id_params#" :: [] in
      Xstring.replaces subst [] text
	
	(** generation of index lemmas and tactics *)
    method index_constructor_lemmas_and_tactics (c : constructor) : (string list * string list) =
      let id_params =
	String.concat " "
	  (List.map (fun p -> self#parameter_rhs_function_gen p) c#get_params) in
      let mod_params =
	String.concat " "
	  (List.map (fun p ->
	    if p#is_index then "(" ^ p#get_name ^ " + _n)"
	    else self#parameter_rhs_function_gen p) c#get_params) in
      let params = String.concat " " (List.map (fun p -> p#get_name) c#get_params) in
      let params_why = String.concat " " (List.map (fun p -> "?" ^ p#get_name) c#get_params) in
      let subst =
	("#lift#", self#fun_name) ::
	("#constr#", c#get_name) ::
	("#index#", c#get_index_name) ::
	("#id_params#", id_params) ::
	("#mod_params#", mod_params) ::
	("#params#", params) ::
	("#why_params#", params_why) ::
	[]
      and text =
	"  Lemma #lift#_#constr#_dec :" ::
	"    forall (_n _m : nat) #params#," ::
	"    { prod (_m <= #index#) (#lift# _n _m (#constr# #params#) = #constr# #mod_params#) } +" ::
	"    { prod (#index# < _m) (#lift# _n _m (#constr# #params#) = #constr# #id_params#) }." ::
	"  Proof." ::
	"    intros; simpl; destruct le_lt_dec; subst; simpl; auto." ::
	"  Qed." ::
	"" ::
	"  Ltac _#lift#_#constr#_dec_tac _n _m #params# :=" ::
	"    let Hcomp := fresh \"Hcomp\" in" ::
	"          let Hfun := fresh \"Hfun\" in" ::
	"            destruct (#lift#_#constr#_dec _n _m #params#) as" ::
	"              [ [ Hcomp Hfun ] | [ Hcomp Hfun ] ];" ::
	"              rewrite ?Hfun in *; clear Hfun." ::
	[]
      and text2 =
	"      | [ |- context [ #lift# ?_n ?_m (#constr# #why_params#) ] ] =>" ::
	"        _#lift#_#constr#_dec_tac _n _m #params#" ::
	"" ::
	"      | [ H : context [ #lift# ?_n ?_m (#constr# #why_params#) ] |- _ ] =>" ::
	"        _#lift#_#constr#_dec_tac _n _m #params#" ::
	"" ::
	[] in
      (Xstring.replaces subst [] text, Xstring.replaces subst [] text2)
	
	(** generation of index equation lemmas *)
    method index_constructor_equation_lemmas (c : constructor) : (string list * string list) =
      let id_params =
	String.concat " "
	  (List.map (fun p -> self#parameter_rhs_function_gen p) c#get_params) in
      let mod_params =
	String.concat " "
	  (List.map (fun p ->
	    if p#is_index then "(" ^ p#get_name ^ " + _n)"
	    else self#parameter_rhs_function_gen p) c#get_params) in
      let params = String.concat " " (List.map (fun p -> p#get_name) c#get_params) in
      let params_why = String.concat " " (List.map (fun p -> "?" ^ p#get_name) c#get_params) in
      let subst =
	("#lift#", self#fun_name) ::
	("#constr#", c#get_name) ::
	("#index#", c#get_index_name) ::
	("#id_params#", id_params) ::
	("#mod_params#", mod_params) ::
	("#params#", params) ::
	("#why_params#", params_why) ::
	[]
      and text =
	"  Lemma #lift#_#constr#_le :" ::
	"    forall (_n _m : nat) #params#," ::
	"      _m <= #index# -> #lift# _n _m (#constr# #params#) = #constr# #mod_params#." ::
	"  Proof." ::
	"    intros; simpl; destruct le_lt_dec; f_equal; omega." ::
	"  Qed." ::
	"" ::
	"  Lemma #lift#_#constr#_lt :" ::
	"    forall (_n _m : nat) #params#," ::
	"      #index# < _m -> #lift# _n _m (#constr# #params#) = #constr# #id_params#." ::
	"  Proof." ::
	"    intros; simpl; destruct le_lt_dec; f_equal; omega." ::
	"  Qed." ::
	[]
      and text2 =
	"      | [ |- context [ #lift# ?_n ?_m (#constr# #why_params#) ] ] =>" ::
	"          (cutrewrite ->" ::
	"            (#lift# _n _m (#constr# #params#) = #constr# #mod_params#);" ::
	"            [ | apply (#lift#_#constr#_le _n _m #params#); cfail ]) ||" ::
	"          (cutrewrite ->" ::
	"            (#lift# _n _m (#constr# #params#) = #constr# #id_params#);" ::
	"            [ | apply (#lift#_#constr#_lt _n _m #params#); cfail ]) ||" ::
	"          #index_dec_tac#" ::
	"" ::
	"      | [ H : context [ #lift# ?_n ?_m (#constr# #why_params#) ] |- _ ] =>" ::
	"          (cutrewrite ->" ::
	"            (#lift# _n _m (#constr# #params#) = #constr# #mod_params#) in H;" ::
	"            [ | apply (#lift#_#constr#_le _n _m #params#); cfail ]) ||" ::
	"          (cutrewrite ->" ::
	"            (#lift# _n _m (#constr# #params#) = #constr# #id_params#) in H;" ::
	"            [ | apply (#lift#_#constr#_lt _n _m #params#); cfail ]) ||" ::
	"          #index_dec_tac#" ::
	"" ::
	[] in
      (Xstring.replaces subst [] text, Xstring.replaces subst [] text2)
  end

class lift_fun_with_source_and_target _gmod _source _target =
  object (self)
    inherit lift_fun _gmod as super
	
    initializer
      begin
	super#set_source _source;
	super#set_target _target
      end
  end

class subst_fun _gmod =
  object (self)
    inherit lift_fun _gmod
	
	(** the definition title *)
    method title : string = "De Bruijn substitution function"
	(** function parameters list (name * type_name) *)
    method fun_params : (string * string) list =
      ("_a", self#get_source#get_name) :: ("_n", "nat") :: []
							    
							    (** radix of the function name *)
    method fun_radix : string = "subst"

    method index_tactic_name : string option = Some "subst_index_tac"
	(** generation of parameters *)
    method private parameter_rhs_function_gen (p : parameter) : string =
      let shift =
	let raw_shifts = List.map self#shift_rhs_function_gen p#get_shifts in
	let non_empty_shifts = List.filter (fun s -> s <> "") raw_shifts in
	String.concat " + " non_empty_shifts in
      let lift_rec =
	p#fold_shift
	  (fun cat -> fun level -> fun r ->
	    if self#get_source = cat
	  || List.mem cat self#get_source#get_throughs then
	      let lift =
		new lift_fun_with_source_and_target gmod cat self#get_source in
	      "(" ^ lift#fun_name ^ " " ^ level ^ " 0 " ^ r ^ ")"
	    else r)
	  "_a" in
      let subst =
	("#param#", p#get_name) ::
	("#recfun#", self#fun_name_type_name p#get_type#get_name) ::
	("#shift#", if shift = "" then "_n" else "(_n + " ^ shift ^ ")") ::
	("#lift_rec#", lift_rec) ::
	[]
      and text =
	if self#is_self_rec p || self#is_back_rec p || self#is_throughs_rec p
	then
	  "(#recfun# #lift_rec# #shift# #param#)"
	else
	  "#param#" in
      Xstring.replace subst text
	
	(** generation of constructors rhs : index case *)
    method private index_constructor_rhs_function_gen (c : constructor) : string list =
      let id_params =
	String.concat " "
	  (List.map (fun p -> self#parameter_rhs_function_gen p) c#get_params) in
      let mod_params =
	String.concat " "
	  (List.map (fun p ->
	    if p#is_index then "(" ^ p#get_name ^ " - 1)"
	    else self#parameter_rhs_function_gen p) c#get_params) in
      let subst =
	("#constr#", c#get_name) ::
	("#index#", c#get_index_name) ::
	("#id_params#", id_params) ::
	("#mod_params#", mod_params) ::
	[]
      and text =
	if self#get_target = self#get_source
	then
	  "match lt_eq_lt_dec _n #index# with" ::
	  "    (* case _n < #index# *)" ::
	  "  | inleft (left _) => #constr# #mod_params#" ::
	  "    (* case _n = #index# *)" ::
	  "  | inleft (right _) => _a" ::
	  "    (* case #index# < _n *)" ::
	  "  | inright _ => #constr# #id_params#" ::
	  "end" ::
	  []
	else "#constr# #index#" :: [] in
      Xstring.replaces subst [] text
	
	(** generation of index lemmas and tactics *)
    method index_constructor_lemmas_and_tactics (c : constructor) : (string list * string list) =
      let id_params =
	String.concat " "
	  (List.map (fun p -> self#parameter_rhs_function_gen p) c#get_params) in
      let mod_params =
	String.concat " "
	  (List.map (fun p ->
	    if p#is_index then "(" ^ p#get_name ^ " - 1)"
	    else self#parameter_rhs_function_gen p) c#get_params) in
      let params = String.concat " " (List.map (fun p -> p#get_name) c#get_params) in
      let params_why = String.concat " " (List.map (fun p -> "?" ^ p#get_name) c#get_params) in
      let subst =
	("#subst#", self#fun_name) ::
	("#constr#", c#get_name) ::
	("#index#", c#get_index_name) ::
	("#id_params#", id_params) ::
	("#mod_params#", mod_params) ::
	("#params#", params) ::
	("#why_params#", params_why) ::
	[]
      and text =
	"  Lemma #subst#_#constr#_dec :" ::
	"    forall (_a : #type#) (_n : nat) #params#," ::
	"    { prod (_n < #index#) (#subst# _a _n (#constr# #params#) = #constr# #mod_params#) } +" ::
	"    { prod (_n = #index#) (#subst# _a _n (#constr# #params#) = _a) } +" ::
	"    { prod (#index# < _n) (#subst# _a _n (#constr# #params#) = #constr# #id_params#) }." ::
	"  Proof." ::
	"    intros; simpl; destruct lt_eq_lt_dec as [ [ ? | ? ] | ? ];" ::
	"      subst; simpl; #crush#." ::
	"  Qed." ::
	"" ::
	"  Ltac _#subst#_#constr#_dec_tac _a _n #params# :=" ::
	"    let Hcomp := fresh \"Hcomp\" in" ::
	"          let Hfun := fresh \"Hfun\" in" ::
	"            destruct (#subst#_#constr#_dec _a _n #params#) as" ::
	"              [ [ [ Hcomp Hfun ] | [ Hcomp Hfun ] ]" ::
	"                | [ Hcomp Hfun ] ];" ::
	"              rewrite ?Hfun in *; clear Hfun." ::
	[]
      and text2 =
	"      | [ |- context [ #subst# ?_a ?_n (#constr# #why_params#) ] ] =>" ::
	"        _#subst#_#constr#_dec_tac _a _n #params#" ::
	"" ::
	"      | [ H : context [ #subst# ?_a ?_n (#constr# #why_params#) ] |- _ ] =>" ::
	"        _#subst#_#constr#_dec_tac _a _n #params#" ::
	"" ::
	[] in
      (Xstring.replaces subst [] text, Xstring.replaces subst [] text2)
	
	(** generation of index equation lemmas *)
    method index_constructor_equation_lemmas (c : constructor) : (string list * string list) =
      let id_params =
	String.concat " "
	  (List.map (fun p -> self#parameter_rhs_function_gen p) c#get_params) in
      let mod_params =
	String.concat " "
	  (List.map (fun p ->
	    if p#is_index then "(" ^ p#get_name ^ " - 1)"
	    else self#parameter_rhs_function_gen p) c#get_params) in
      let params = String.concat " " (List.map (fun p -> p#get_name) c#get_params) in
      let params_why = String.concat " " (List.map (fun p -> "?" ^ p#get_name) c#get_params) in
      let subst =
	("#subst#", self#fun_name) ::
	("#constr#",  c#get_name) ::
	("#index#", c#get_index_name) ::
	("#id_params#", id_params) ::
	("#mod_params#", mod_params) ::
	("#params#", params) ::
	("#why_params#", params_why) ::
	[]
      and text =
	"  Lemma #subst#_#constr#_lt_l :" ::
	"    forall (_a : #type#) (_n : nat) #params#," ::
	"      _n < #index# -> #subst# _a _n (#constr# #params#) = #constr# #mod_params#." ::
	"  Proof." ::
	"    intros; simpl; destruct lt_eq_lt_dec as [ [ ? | ? ] | ? ]; #crush#." ::
	"  Qed." ::
	"" ::
	"  Lemma #subst#_#constr#_eq :" ::
	"    forall (_a : #type#) (_n : nat) #params#," ::
	"      _n = #index# -> #subst# _a _n (#constr# #params#) = _a." ::
	"  Proof." ::
	"    intros; simpl; destruct lt_eq_lt_dec as [ [ ? | ? ] | ? ]; #crush#." ::
	"  Qed." ::
	"" ::
	"  Lemma #subst#_#constr#_lt_r :" ::
	"    forall (_a : #type#) (_n : nat) #params#," ::
	"      #index# < _n -> #subst# _a _n (#constr# #params#) = #constr# #id_params#." ::
	"  Proof." ::
	"    intros; simpl; destruct lt_eq_lt_dec as [ [ ? | ? ] | ? ]; #crush#." ::
	"  Qed." ::
	[]
      and text2 =
	"        | [ |- context [ #subst# ?_a ?_n (#constr# #why_params#) ] ] =>" ::
	"          (cutrewrite ->" ::
	"            (#subst# _a _n (#constr# #params#) = #constr# #mod_params#);" ::
	"            [ | apply (#subst#_#constr#_lt_l _a _n #params#); cfail ]) ||" ::
	"          (cutrewrite ->" ::
	"            (#subst# _a _n (#constr# #params#) = _a);" ::
	"            [ | apply (#subst#_#constr#_eq _a _n #params#); cfail ]) ||" ::
	"          (cutrewrite ->" ::
	"            (#subst# _a _n (#constr# #params#) = #constr# #id_params#);" ::
	"            [ | apply (#subst#_#constr#_lt_r _a _n #params#); cfail ]) ||" ::
	"          #index_dec_tac#" ::
	"" ::
	"        | [ H : context [ #subst# ?_a ?_n (#constr# #why_params#) ] |- _ ] =>" ::
	"          (cutrewrite ->" ::
	"            (#subst# _a _n (#constr# #params#) = #constr# #mod_params#) in H;" ::
	"            [ | apply (#subst#_#constr#_lt_l _a _n #params#); cfail ]) ||" ::
	"          (cutrewrite ->" ::
	"            (#subst# _a _n (#constr# #params#) = _a) in H;" ::
	"            [ | apply (#subst#_#constr#_eq _a _n #params#); cfail ]) ||" ::
	"          (cutrewrite ->" ::
	"            (#subst# _a _n (#constr# #params#) = #constr# #id_params#) in H;" ::
	"            [ | apply (#subst#_#constr#_lt_r _a _n #params#); cfail ]) ||" ::
	"          #index_dec_tac#" ::
	"" ::
	[] in
      (Xstring.replaces subst [] text, Xstring.replaces subst [] text2)
  end

class subst_fun_with_source_and_target _gmod _source _target =
  object
    inherit subst_fun _gmod as super
	
    initializer
      begin
	super#set_source _source;
	super#set_target _target
      end
  end
