open Debug

open Xlist
open Xstring
open Xcoq

open Interfaces

open Syntax_module_template

open Definitions
open Properties

(** class of modules, ordered list of nodes *)
class gmodule_impl _name _coq_predefs _nodes _coq_postdefs : gmodule =
  object (self)
      (** name of the module *)
    val name : string = _name
	(** verbatim preliminary definitions in coq *)
    val coq_predefs : string list = _coq_predefs
	(** list of elements *)
    val nodes : node list = _nodes
	(** verbatim appendix definitions in coq *)
    val coq_postdefs : string list = _coq_postdefs
	
	(** initializer : give reference to this gmodule to every sub-element *)
    initializer
      if nodes <> [] then
	begin
	  Debug.println "### gmodule initializer started.";
	  List.iter (fun (n : node) -> n#set_gmodule (self :> gmodule)) nodes;
	  Debug.println "### gmodule self linking done.";
	  self#throughs_init;
	  Debug.println "### grammars throughs initialization done.";
	  self#throughs_expand;
	  Debug.println "### grammars throughs expansion done.";
	  Debug.print (self#to_string())
	end
      else ()
	
	(** returns all the grammars of the gmodule *)
    method get_all_grams : grammar list =
      List.concat
	(List.map (fun n -> n#get_grams) nodes)
	
	(** returns all the grammars that has index in the gmodule *)
    method get_indexed_grams : grammar list =
      List.filter (fun g -> g#has_index) self#get_all_grams
	
	(** returns all the grammars that has index in the gmodule and
	 * that can be reached from the grammar of the given name *)
    method get_indexed_grams_through (gr : grammar) : grammar list =
      List.filter
	(fun g -> g = gr || List.mem g gr#get_throughs)
	self#get_indexed_grams
	
	(** returns the grammar of the given name as an option *)
    method get_gram (gname : string) : grammar =
      List.find (fun g -> g#get_name = gname) self#get_all_grams
	
	(** returns true/false whether the grammar is one of the gmodule *)
    method is_gmodule_gram (gname : string) : bool =
      try (let _ = self#get_gram gname in true)
      with Not_found -> false
	  
	  (** initializes the through list, collecting immediate
	   * subcategories that has index for every grammar of this module *)
    method private throughs_init : unit =
      List.iter (fun n -> n#throughs_init) nodes
	
	(** expands the through list, adding immediate
	 * subcategories throughs *)
    method private throughs_expand : unit =
      (* to avoid infinite loop when debugging, we maximize the number of recursive calls *)
      let max_expand_count = 100 in
      let rec aux nb =
	if nb = max_expand_count then
	  begin
	    Debug.println "Ast.gmodule#throughs_expand : max_expand_count reached, exiting.";
	    exit 1
	  end else
	  let result =
	    List.fold_right
	      (fun n -> fun result -> n#throughs_expand || result)
	      nodes false
	  in if result then aux (nb + 1) else ()
      in aux 0
	
	(** syntax definition generation *)
    method private syntax_gen (def : terms_definition) : string list =
      let _ = Debug.println ("#### Syntax generation: " ^ def#title) in
      def#coq_predefs @ [""] @
      ("  (** *** " ^ def#title ^ " *)") :: [""] @
      (Xlist.concat_with
	 [""]
	 (List.map (fun n -> n#syntax_gen def) nodes))
	
	(** syntax definition generation *)
    method private function_gen (def : fun_definition) : string list =
      let _ = Debug.println ("#### Function generation: " ^ def#title) in
      let one_source_gen g =
	let _ =
	  match g with
	    Some gram -> def#set_source gram
	  | None -> ()
	in
	def#coq_predefs @ [""] @
	("  (** *** " ^ def#title ^ " *)") :: [""] @
	(Xlist.concat_with
	   [""]
	   (List.map (fun n -> n#function_gen def) nodes))
      in
      List.concat
	(List.map one_source_gen def#get_source_grams)
	
	(** terms and functions definitions for a given template list *)
    method private terms_and_funs_def : string list =
      let rec aux l =
	match l with
	  [] -> []
	| Term def :: l ->
	    let gen = self#syntax_gen def in
	    gen @ [""] @ aux l
	| Fun def :: l ->
	    let gen = self#function_gen def in
	    gen @ [""] @ aux l
      in
      aux (Definitions.ordered_defs (self :> gmodule))
	
	(** property generation *)
    method property_gen (prop : property)
	: string list * (string * (string list * string list)) list =
      let (lemmas, tactics) =
	List.split
	  (List.map (fun n -> n#property_gen prop) nodes) in
      (Xlist.concat_with [""] lemmas, List.concat tactics)
	
	(** properties generation for all chapters *)
    method private all_properties_gen : string list =
      let rec aux l =
	match l with
	  [] -> []
	| chap :: l -> chap#generate @ [""] @ aux l in
      aux (Properties.ordered_chaps (self :> gmodule))

    method private all_properties_tac : string list =
      let tactics =
	String.concat " || "
	  ("#index_tac#" ::
	   Xlist.option_prune
	     (List.map (fun chap -> chap#tactic_name)
		(Properties.ordered_chaps (self :> gmodule)))) in
      if self#get_indexed_grams <> [] then
	("  Ltac dbgen_tac :=") ::
	("    #crush#;") ::
	("    repeat (progress (" ^ tactics ^ ")).") ::
	[]
      else []
	
	(** coq output generation *)
    method coq_output : string list =
      (** Global names used in all definitions and properties *)
      let subst =
	("#gmodule#", name) ::
	("#database#", name ^ "_database") ::
	("#crush#", "crush_tac") ::
	("#ecrush#", "ecrush_tac") ::
	("#index_tac#", "index_tac") ::
	[] in
      
      (* Coq module header and footer *)
      let header = SyntaxModuleTemplate.header
      and footer = SyntaxModuleTemplate.footer in
      let _ = Debug.println "### Module header and footer invoked." in
      
      let title_def title def =
	Xcoq.coq_title 1 title @ [""] @ def @ [""] in
      
      let terms_and_funs_def = self#terms_and_funs_def in
      let _ = Debug.println "### Terms and function successfully generated." in
      
      let all_properties = self#all_properties_gen in
      let _ = Debug.println "### Properties successfully generated." in

      let all_tactics = self#all_properties_tac in
      let _ = Debug.println "### Tactics successfully generated." in
      
      let output =
	(* header and user pre-definitions *)
	header @ [""] @
	coq_predefs @ [""] @
	
	(title_def "Inductive syntax and function definitions"
	   terms_and_funs_def) @
	
	(title_def "Properties"
	   all_properties) @
	
	(title_def "Main dbgen tactic"
	   all_tactics) @
	
	(* user post-definitions and footer *)
	coq_postdefs @ [""] @
	footer in
      Xstring.replaces subst [] output
	
	(** Pretty printing *)
    method to_string () : string =
      "Module " ^ name ^ ".\n\n" ^
      (String.concat "\n" coq_predefs) ^ "\n" ^
      (String.concat
	 "\n"
	 (List.map (fun n -> n#to_string ()) nodes)) ^
      (String.concat "\n" coq_postdefs) ^ "\n" ^
      "\nEnd " ^ name ^ ".\n"
  end
