open Error
open Xlist

open Interfaces

(** classe of nodes, i.e. collection of mutually defined grammars *)
class node_impl _gmod _grams _reg : node =
  object (self)
      (** grammar module of this element *)
    val mutable gmod : gmodule = _gmod
	(** position in the source file *)
    val reg : Error.region = _reg

	(** list of grammar of this node *)
    val grams : grammar list = _grams
	
	(** returns the grammar list of this node *)
    method get_grams : grammar list = grams
	
	(** post-initialization with the given gmodule *)
    method set_gmodule (gm : gmodule) : unit =
      begin
	(* set the gmod field *)
	gmod <- gm;
	(* performs post_initialization for grammars *)
	List.iter (fun g -> g#set_gmodule gm) grams
      end
	
	(** initialize the through list of each grammar *)
    method throughs_init : unit =
      List.iter (fun g -> g#throughs_init) grams
	
	(** expand the through list of each grammar *)
    method throughs_expand : bool =
      List.fold_right
	(fun g -> fun result -> g#throughs_expand || result) grams false
	
	(** syntax definition generation *)
    method syntax_gen (def : terms_definition) : string list =
      (Xlist.concat_with [""]
	 (Xlist.map_with_hd_tl_param "Inductive" "with"
	    (fun prefix -> fun g -> def#grammar_terms_gen prefix g)
	    grams)) @
      "  ." ::
      "" ::
      ("  Hint Constructors "
       ^ String.concat " " (List.map (fun g -> def#decorate_type_name g#get_name) grams)
       ^ " : #database#.") ::
      []
	
	(** function definition generation *)
    method function_gen (def : fun_definition) : string list =
      let target_grams = def#get_target_grams grams in
      let fun_def =
	(Xlist.concat_with [""]
	   (Xlist.map_with_hd_tl_param "Fixpoint" "with"
	      (fun prefix -> fun g -> def#grammar_function_gen prefix g)
	      target_grams)) @
	"  ." ::
	"" ::
	(Xlist.concat_with [""] (List.map (fun g -> g#fun_equations_gen def) grams)) in
      if target_grams = [] then []
      else fun_def
	  
	  (** property generation *)
    method property_gen (prop : property)
	: string list * (string * (string list * string list)) list =
      let target_grams = prop#get_target_grams grams in
      let results =
	Xlist.map_with_hd_tl_param "Lemma" "with"
	  (fun prefix -> fun g -> g#property_gen prop prefix (self :> node))
	  target_grams in
      let (lemmas, proofs, aux_text, tactics) = Xlist.split_4 results in
      let all_lemmas =
	(Xlist.concat_with [""] lemmas) @
	"  ." ::
	"  Proof." ::
	(Xlist.concat_with [""] proofs) @
	"  Qed." ::
	"" ::
	(Xlist.concat_with [""] aux_text) in
      if target_grams = [] then ([], [])
      else (all_lemmas, tactics)
	  
	  (** Pretty printing *)
    method to_string () : string =
      "  Inductive " ^
      (String.concat
	 "\n\n  with " (List.map (fun g -> g#to_string ()) grams)) ^
      ".\n"
  end
