open Error
open Xlist

open Interfaces

(** class of constructor parameters *)
class parameter_impl _gmod _shifts _name _gtype _reg : parameter =
  object
      (** grammar module of this element *)
    val mutable gmod : gmodule = _gmod
	(** position in the source file *)
    val reg : Error.region = _reg

	(** binding categories (type_name, shift_level) *)
    val mutable shifts : (grammar * shift_exp) list =
      (* we aggregate all the shift on the same category since
	 interleaving with other categories shift is irrelevant *)
      Xlist.assoc_aggregate plusify _shifts
	(** parameter name used for pretty printing *)
    val name : string = _name
	(** category name of the parameter *)
    val mutable gtype : grammar = _gtype
	
	(** accessors *)
    method get_name : string = name
    method get_type : grammar = gtype
    method get_shifts : (grammar * shift_exp) list = shifts
	
	(** returns true/false whether this is an index parameter *)
    method is_index : bool = false
	
	(** post-initialization with the given gmodule *)
    method set_gmodule (gm : gmodule) : unit =
      (* returns the gmodule grammar replacing the given shifted grammar,
	 fails if undefined or without index *)
      let get_gmodule_gram g =
	try
	  let new_g = gm#get_gram g#get_name in
	  if new_g#has_index then new_g else
	  Error.fail reg ("no index found in binding category " ^ new_g#get_name)
	with Not_found ->
	  Error.fail reg ("undefined binding category " ^ g#get_name) in
      begin
	(* set the gmod field *)
	gmod <- gm;
	(* connect gtype and shifted grammars *)
	gtype <- (try gm#get_gram gtype#get_name with Not_found -> gtype);
	shifts <- List.map (fun (g, n) -> (get_gmodule_gram g, n)) shifts
      end
	
	(** returns the through list of the category, i.e.
	 * the category name is it has index, and the throughs list
	 * of the category *)
    method get_throughs : grammar list =
      (if gtype#has_index then [gtype] else []) @
      gtype#get_throughs
	
	(** induction pattern for this parameter *)
    method induction_pattern (g : grammar) : string list =
      name :: if gtype = g then [ "IH" ^ name ] else []
	
	(** rewrite induction hypothesis for this parameter *)
    method rewrite_induction (prop : property) (source : grammar) : string =
      if gtype = source
      then "try rewrite IH" ^ name ^ "; #crush#"
      else
	if List.mem gtype (prop#get_target_grams gmod#get_all_grams)
	then "try rewrite " ^ prop#lemma_name_type_name gtype#get_name ^ "; #crush#"
	else "#crush#"
	    
	    (** fold a function on the shift list *)
    method fold_shift (f : grammar -> string -> string -> string) (init : string) : string =
      List.fold_right
	(fun (cat, level) -> fun r -> f cat (string_of_shift level) r)
	shifts init
	
	(** Pretty printing *)
    method to_string () : string =
      " (" ^
      (if List.length shifts = 0 then "" else
      "(* bind " ^
      (String.concat ", " (List.map (fun (g, n) -> string_of_shift n ^ " " ^ g#get_name) shifts)) ^
      " in *) ") ^
      name ^ " : " ^ gtype#get_name ^
      ")"
  end

(** class of index constructor parameters *)
class index_parameter_impl _gmod _name _gtype _reg : parameter =
  object
    inherit parameter_impl _gmod [] _name _gtype _reg
	
	(** returns true/false whether this is an index parameter *)
    method is_index : bool = true
	
	(** rewrite induction hypothesis for this parameter *)
    method rewrite_induction (prop : property) (g : grammar) : string =
      prop#get_index_tac ^ "; #crush#"
			     
			     (** Pretty printing *)
    method to_string () : string =
      "((* index *) " ^ name ^ " : " ^ gtype#get_name ^ ")"
  end
