open Xlist

open Interfaces
open Properties_abstract
open Definitions_lift_subst

class virtual syntax_directed_property _gmod =
  object (self)
    inherit virtual_property _gmod as super
	
	(** the source grammars for which generation is needed.
	 * In this case : every indexed grammar. *)
    method get_source_grams : (grammar * grammar option) list = []
	
	(** the target grammars to process among a given grammar list *)
    method get_target_grams (grams : grammar list) : grammar list = grams
	
    method get_index_tac : string = "#index_tac#; simpl"
  end

class virtual indexed_double_ordered_directed_property _gmod =
  object (self)
    inherit virtual_property _gmod as super
	
	(** the source grammars for which generation is needed.
	 * In this case : every indexed grammar. *)
    method get_source_grams : (grammar * grammar option) list =
      List.map (fun (g1, g2) -> (g1, Some g2))
	(Xlist.build_ordered_combin gmod#get_indexed_grams)
	
	(** the target grammars to process among a given grammar list *)
    method get_target_grams (grams : grammar list) : grammar list =
      List.filter
	(fun g ->
	  (List.mem self#get_source (g :: g#get_throughs)) &&
	  (List.mem self#get_alt_source (g :: g#get_throughs)))
	grams
  end

class virtual indexed_double_directed_property _gmod =
  object (self)
    inherit virtual_property _gmod as super
	
	(** the source grammars for which generation is needed.
	 * In this case : every indexed grammar. *)
    method get_source_grams : (grammar * grammar option) list =
      List.map (fun (g1, g2) -> (g1, Some g2))
	(Xlist.build_combin gmod#get_indexed_grams)
	
	(** the target grammars to process among a given grammar list *)
    method get_target_grams (grams : grammar list) : grammar list =
      List.filter
	(fun g ->
	  (List.mem self#get_source (g :: g#get_throughs)) &&
	  (List.mem self#get_alt_source (g :: g#get_throughs)))
	grams
  end

(**
 * lifting independance.
 *)
class lift_indep_property _gmod =
  object (self)
    inherit indexed_double_ordered_directed_property _gmod as super
	
    method lift_1 : lift_fun =
      new Definitions_lift_subst.lift_fun_with_source_and_target
	gmod self#get_source self#get_target
	
    method lift_2 : lift_fun =
      new Definitions_lift_subst.lift_fun_with_source_and_target
	gmod self#get_alt_source self#get_target
	
    method subst : (string * string) list =
      ("#lift_1#", self#lift_1#fun_name) ::
      ("#lift_2#", self#lift_2#fun_name) ::
      []
	
	(** name of the lemma *)
    method lemma_name_type_name (name : string) : string =
      self#lift_1#fun_name_type_name name ^ "_" ^
      self#lift_2#fun_name_type_name name ^ "_indep"
					    
					    (** the lemma statement *)
    method lemma_header : string list =
      "  #Lemma# #lemma_name# :" ::
      "    forall (_arg : #target_type#) (_n _m _n' _m' : nat)," ::
      "      #lift_1# _n _m (#lift_2# _n' _m' _arg) = " ::
      "      #lift_2# _n' _m' (#lift_1# _n _m _arg)" ::
      []
	
	(** the lemma proof script *)
    method lemma_proof : string list =
      "    #by_induction#;" ::
      "      intros; " ::
      "#rewrite_induction_hypothesis#" ::
      "      idtac." ::
      []
	
	(** extra definition lines to be inserted after the property *)
    method aux_defs : string list =
      "  Hint Rewrite #lemma_name# : #database#." ::
      []
	
    method get_index_tac : string = "#index_tac#; simpl; try destruct le_lt_dec as [ ? | ? ]"
	
	(** the tactic template : pattern, body and body postpone *)
    method tactic_template : string * (string list * string list) =
      ("", ([], []))
  end

(**
 * lifting-substitution independance.
 *)
class lift_subst_indep_property _gmod =
  object (self)
    inherit indexed_double_directed_property _gmod as super
	
    method lift_target : lift_fun =
      new Definitions_lift_subst.lift_fun_with_source_and_target
	gmod self#get_source self#get_target
	
    method lift_source : lift_fun =
      new Definitions_lift_subst.lift_fun_with_source_and_target
	gmod self#get_source self#get_alt_source
	
    method subst_target : lift_fun =
      new Definitions_lift_subst.subst_fun_with_source_and_target
	gmod self#get_alt_source self#get_target
	
    method subst : (string * string) list =
      ("#alt_source_type#", self#get_alt_source#get_name) ::
      ("#lift_target#", self#lift_target#fun_name) ::
      ("#lift_source#", self#lift_source#fun_name) ::
      ("#subst_target#", self#subst_target#fun_name) ::
      []
	
	(** name of the lemma *)
    method lemma_name_type_name (name : string) : string =
      self#lift_target#fun_name_type_name name ^ "_" ^
      self#subst_target#fun_name_type_name name ^ "_indep"
					    
					    (** the lemma statement *)
    method lemma_header : string list =
      let new_arg' =
	if List.mem self#get_source self#get_alt_source#get_throughs
	then "(#lift_source# _n _m _arg')"
	else "_arg'" in
      "  #Lemma# #lemma_name# :" ::
      "    forall (_arg : #target_type#) (_n _m _p : nat) (_arg' : #alt_source_type#)," ::
      "      #lift_target# _n _m (#subst_target# _arg' _p _arg) = " ::
      ("      #subst_target# " ^ new_arg' ^ " _p (#lift_target# _n _m _arg)") ::
      []
	
	(** the lemma proof script *)
    method lemma_proof : string list =
      "    #by_induction#;" ::
      "      intros; " ::
      "#rewrite_induction_hypothesis#" ::
      "      lift_lift_tac; crush_tac." ::
      []
	
	(** extra definition lines to be inserted after the property *)
    method aux_defs : string list =
      "  Hint Rewrite #lemma_name# : #database#." ::
      []
	
    method get_index_tac : string = "#index_tac#; simpl; try destruct lt_eq_lt_dec as [ [ ? | ? ] | ? ]; try destruct le_lt_dec as [ ? | ? ]"
	
	(** the tactic template : pattern, body and body postpone *)
    method tactic_template : string * (string list * string list) =
      ("", ([], []))
  end

class indep_pack _gmod =
  object
    inherit virtual_property_pack _gmod
	
	(** the title to insert before the definitions *)
    method get_title : string =
      "Lifting and substitution respective independance"
	
	(** the list of properties of this pack *)
    method properties : property list =
      (new lift_indep_property gmod :> property) ::
      (new lift_subst_indep_property gmod :> property) ::
      []
  end

(***** Indexed directed properties *)

class virtual indexed_directed_property _gmod =
  object (self)
    inherit virtual_property _gmod as super
	
	(** the source grammars for which generation is needed.
	 * In this case : every indexed grammar. *)
    method get_source_grams : (grammar * grammar option) list =
      List.map (fun g -> (g, None)) gmod#get_indexed_grams
	
	(** the target grammars to process among a given grammar list *)
    method get_target_grams (grams : grammar list) : grammar list =
      List.filter
	(fun g ->
	  List.mem self#get_source (g :: g#get_throughs))
	grams
	
    method get_index_tac : string = "#index_tac#; simpl"
  end

class virtual lift_property _gmod =
  object (self)
    inherit indexed_directed_property _gmod as super
	
    method lift : lift_fun =
      new Definitions_lift_subst.lift_fun_with_source_and_target
	gmod self#get_source self#get_target
	
    method subst : (string * string) list =
      ("#lift#", self#lift#fun_name) ::
      []
  end

(* TODO problem with mixed existential induction.

(**
 * lifting functional inversion.
 *)
  val liftInv_template : lemmaTemplate =
  let val induction_pattern = []
      (try_rew_mix g) @
    mutual_invokes
  in
  {
  lemma_nameGen =
  fn (gmain : grammar) => fn (g : grammar) =>
  ((#fun_name lift_template) (SOME gmain) g) ^ "_fun_inv",
  lemma_header =
  "    forall ($lp1 $lp2 $lp1' $lp2' : nat) (arg1 arg2 : $type)," ::
  "      $lp1' = $lp1 -> $lp2' = $lp2 ->" ::
  "      $lift $lp1 $lp2 arg1 = $lift $lp1' $lp2' arg2 ->" ::
  "      arg1 = arg2" ::
  [],
  lemma_proof =
  "    intros $lp1 $lp2 $lp1' $lp2' arg1 arg2 Heq1 Heq2; subst;" ::
  "      generalize arg1 arg2 $lp1 $lp2; clear arg1 arg2 $lp1 $lp2." ::
  "    intro arg1; $induction arg1; intro arg2; $induction' arg2;" ::
  "      intros $lp1 $lp2 Heq; $vartac;" ::
  "        simpl in Heq; inversion Heq; subst; $crush;" ::
  induction_pattern @
  "        $crush." ::
  [],
  aux_defs =
  "  Hint Rewrite $lname : $database." ::
  [],
  tactic_pattern = "",
  tactic_body = [],
  tactic_postpone = []
  }
  end
 *)


(**
 * lifting with 0 simplification.
 *)
class lift_0_simpl_property _gmod =
  object (self)
    inherit lift_property _gmod as super
	
	(** name of the lemma *)
    method lemma_name_type_name (name : string) : string =
      self#lift#fun_name_type_name name ^ "_0_simpl"
					    
					    (** the lemma statement *)
    method lemma_header : string list =
      "  #Lemma# #lemma_name# :" ::
      "    forall (_arg : #target_type#) (_m : nat)," ::
      "      #lift# 0 _m _arg = _arg" ::
      []
	
	(** the lemma proof script *)
    method lemma_proof : string list =
      "    #by_induction#;" ::
      "      intro _m; " ::
      "#rewrite_induction_hypothesis#" ::
      "      idtac." ::
      []
	
	(** extra definition lines to be inserted after the property *)
    method aux_defs : string list =
      "  Hint Rewrite #lemma_name# : #database#." ::
      []
	
	(** the tactic template : pattern, body and body postpone *)
    method tactic_template : string * (string list * string list) =
      ("", ([], []))
  end

class lift_0_simpl_pack _gmod =
  object
    inherit virtual_property_pack _gmod
	
	(** the title to insert before the definitions *)
    method get_title : string =
      "Lifting with 0 simplification"
	
	(** the list of properties of this pack *)
    method properties : property list =
      (new lift_0_simpl_property gmod :> property) ::
      []
  end

(**
 * lifting with 0 simplification generalization.
 *)
class lift_0_simpl_gen_property _gmod =
  object (self)
    inherit lift_property _gmod as super
	
	(** name of the lemma *)
    method lemma_name_type_name (name : string) : string =
      self#lift#fun_name_type_name name ^ "_0_simpl_gen"
					    
					    (** the lemma statement *)
    method lemma_header : string list =
      "  #Lemma# #lemma_name# :" ::
      "    forall (_arg : #target_type#) (_n _m : nat)," ::
      "      _n = 0 -> #lift# _n _m _arg = _arg" ::
      []
	
	(** the lemma proof script *)
    method lemma_proof : string list =
      "    intros; subst; #crush#." ::
      []
	
	(** extra definition lines to be inserted after the property *)
    method aux_defs : string list =
      []
	
	(** the tactic template : pattern, body and body postpone *)
    method tactic_template : string * (string list * string list) =
      ("#lift# ?n ?m ?t",
       ("          (cutrewrite ->" ::
	"            (#lift# n m t = t)#inH#;" ::
	"            [ | apply (#lemma_name# t n m); cfail ])" ::
	[], []))
  end

class lift_0_simpl_gen_pack _gmod =
  object
    inherit virtual_property_pack _gmod
	
	(** the title to insert before the definitions *)
    method get_title : string =
      "Lifting with 0 simplification generalized"
	
	(** the list of properties of this pack *)
    method properties : property list =
      (new lift_0_simpl_gen_property gmod :> property) ::
      []
	
	(** the tactic name of the tactic associated with this pack,
	 * or None if no tactic is required *)
    method tactic_name : string option = Some "lift_0_simpl_tac"
  end

(**
 * lifting over lifting simplification.
 *)
class lift_lift_simpl_property _gmod =
  object (self)
    inherit lift_property _gmod as super
	
	(** name of the lemma *)
    method lemma_name_type_name (name : string) : string =
      self#lift#fun_name_type_name name ^ "_" ^
      self#lift#fun_name_type_name name ^ "_simpl"
					    
					    (** the lemma statement *)
    method lemma_header : string list =
      "  #Lemma# #lemma_name# :" ::
      "    forall (_arg : #target_type#) (_n _m _n' _m' : nat)," ::
      "      _m' <= _m <= _n' + _m' ->" ::
      "      #lift# _n _m (#lift# _n' _m' _arg) =" ::
      "      #lift# (_n' + _n) _m' _arg" ::
      []
	
	(** the lemma proof script *)
    method lemma_proof : string list =
      "    #by_induction#;" ::
      "      intros _n _m _n' _m' Hyp; destruct Hyp;" ::
      "#rewrite_induction_hypothesis#" ::
      "      idtac." ::
      []
	
	(** extra definition lines to be inserted after the property *)
    method aux_defs : string list =
      []
	
	(** the tactic template : pattern, body and body postpone *)
    method tactic_template : string * (string list * string list) =
      ("#lift# ?n ?m (#lift# ?p ?q ?t)",
       ("          (cutrewrite ->" ::
	"            (#lift# n m (#lift# p q t) =" ::
	"             #lift# (p + n) q t)#inH#;" ::
	"            [ | apply (#lemma_name# t n m p q); cfail ])" ::
	[], []))
  end

(**
 * lifting over lifting commutativity.
 *)
class lift_lift_commut_property _gmod =
  object (self)
    inherit lift_property _gmod as super
	
	(** name of the lemma *)
    method lemma_name_type_name (name : string) : string =
      self#lift#fun_name_type_name name ^ "_" ^
      self#lift#fun_name_type_name name ^ "_commut"
					    
					    (** the lemma statement *)
    method lemma_header : string list =
      "  #Lemma# #lemma_name# :" ::
      "    forall (_arg : #target_type#) (_n _m _n' _m' : nat)," ::
      "      _n' + _m' <= _m ->" ::
      "      #lift# _n _m (#lift# _n' _m' _arg) =" ::
      "      #lift# _n' _m' (#lift# _n (_m - _n') _arg)" ::
      []
	
	(** the lemma proof script *)
    method lemma_proof : string list =
      "    #by_induction#; intros;" ::
      "#rewrite_induction_hypothesis#" ::
      "      idtac." ::
      []
	
	(** extra definition lines to be inserted after the property *)
    method aux_defs : string list =
      []
	
	(** the tactic template : pattern, body and body postpone *)
    method tactic_template : string * (string list * string list) =
      ("#lift# ?n ?m (#lift# ?p ?q ?t)",
       ("          (cutrewrite ->" ::
	"            (#lift# n m (#lift# p q t) =" ::
	"             #lift# p q (#lift# n (m - p) t))#inH#;" ::
	"            [ | apply (#lemma_name# t n m p q); cfail ])" ::
	[], []))
  end

class lift_lift_pack _gmod =
  object
    inherit virtual_property_pack _gmod
	
	(** the title to insert before the definitions *)
    method get_title : string =
      "Lifting over lifting simplification and commutation"
	
	(** the list of properties of this pack *)
    method properties : property list =
      (new lift_lift_simpl_property gmod :> property) ::
      (new lift_lift_commut_property gmod :> property) ::
      []
	
	(** the tactic name of the tactic associated with this pack,
	 * or None if no tactic is required *)
    method tactic_name : string option = Some "lift_lift_tac"
  end

class lift_chapter _gmod =
  object
    inherit virtual_property_chapter _gmod
	
	(** the title to insert before the definitions *)
    method get_title : string =
      "Lifting main properties"
	
	(** index lemmas *)
    method fun_index_lemmas_and_tactics_to_generate : fun_definition list =
      (new lift_fun gmod) ::
      (new subst_fun gmod) ::
      []
	
	(** the index tactic name of the tactic associated with this chapter,
	 * or None if no tactic is required *)
    method index_tactic_name : string option = Some "index_tac"
	
	(** the tactic name of the tactic associated with this chapter,
	 * or None if no tactic is required *)
    method tactic_name : string option = Some "lift_tac"
    method aux_tactics : string list =
      "lift_index_tac" ::
      []

	(** the list of property packs of this chapter *)
    method property_packs : property_pack list =
      (new lift_0_simpl_pack gmod :> property_pack) ::
      (new lift_0_simpl_gen_pack gmod :> property_pack) ::
      (new lift_lift_pack gmod :> property_pack) ::
      []
  end

class virtual subst_property _gmod =
  object (self)
    inherit lift_property _gmod as super
	
    method subst_f : subst_fun =
      new Definitions_lift_subst.subst_fun_with_source_and_target
	gmod self#get_source self#get_target
	
    method subst : (string * string) list =
      ("#subst#", self#subst_f#fun_name) ::
      super#subst
  end

class subst_lift_simpl_property _gmod =
  object (self)
    inherit subst_property _gmod as super
	
	(** name of the lemma *)
    method lemma_name_type_name (name : string) : string =
      self#subst_f#fun_name_type_name name ^ "_" ^
      self#lift#fun_name_type_name name ^ "_simpl"
					    
					    (** the lemma statement *)
    method lemma_header : string list =
      "  #Lemma# #lemma_name# :" ::
      "    forall (_arg : #target_type#) (_arg' : #source_type#) (_n _m _p : nat)," ::
      "      _m <= _p < _m + _n ->" ::
      "      #subst# _arg' _p (#lift# _n _m _arg) =" ::
      "      #lift# (_n - 1) _m _arg" ::
      []
	
	(** the lemma proof script *)
    method lemma_proof : string list =
      "    #by_induction#; intros;" ::
      "#rewrite_induction_hypothesis#" ::
      "      idtac." ::
      []
	
	(** extra definition lines to be inserted after the property *)
    method aux_defs : string list =
      []
	
	(** the tactic template : pattern, body and body postpone *)
    method tactic_template : string * (string list * string list) =
      ("#subst# ?v ?p (#lift# ?n ?m ?t)",
       ("          (cutrewrite ->" ::
	"            (#subst# v p (#lift# n m t) =" ::
	"             #lift# (n - 1) m t)#inH#;" ::
	"            [ | apply (#lemma_name# t v n m p); cfail ])" ::
	[], []))
  end

class subst_lift_simpl_0_property _gmod =
  object (self)
    inherit subst_property _gmod as super
	
	(** name of the lemma *)
    method lemma_name_type_name (name : string) : string =
      self#subst_f#fun_name_type_name name ^ "_" ^
      self#lift#fun_name_type_name name ^ "_0_simpl"
					    
					    (** the lemma statement *)
    method lemma_header : string list =
      "  #Lemma# #lemma_name# :" ::
      "    forall (_arg : #target_type#) (_n _m : nat)," ::
      "      0 < _n ->" ::
      "      #subst# (#index_constr# _m) _m (#lift# _n (S _m) _arg) =" ::
      "      #lift# (_n - 1) (S _m) _arg" ::
      []
	
	(** the lemma proof script *)
    method lemma_proof : string list =
      "    #by_induction#; intros;" ::
      "#rewrite_induction_hypothesis#" ::
      "      #crush#." ::
      []
	
	(** extra definition lines to be inserted after the property *)
    method aux_defs : string list =
      []
	
	(** the tactic template : pattern, body and body postpone *)
    method tactic_template : string * (string list * string list) =
      ("#subst# (#index_constr# ?m) ?m (#lift# ?n (S ?m) ?t)",
       ("          (cutrewrite ->" ::
	"            (#subst# (#index_constr# m) m (#lift# n (S m) t) =" ::
	"             #lift# (n - 1) (S m) t)#inH#;" ::
	"            [ | apply (#lemma_name# t n m); cfail ])" ::
	[], []))
  end

class subst_lift_commut_property _gmod =
  object (self)
    inherit subst_property _gmod as super
	
	(** name of the lemma *)
    method lemma_name_type_name (name : string) : string =
      self#subst_f#fun_name_type_name name ^ "_" ^
      self#lift#fun_name_type_name name ^ "_commut"
					    
    method subst : (string * string) list =
      ("#source_lift#", self#lift#fun_name_type_name self#get_source#get_name) ::
      super#subst

	(** the lemma statement *)
    method lemma_header : string list =
      "  #Lemma# #lemma_name# :" ::
      "    forall (_arg : #target_type#) (_arg' : #source_type#) (_n _m _p : nat)," ::
      "      _m <= _p ->" ::
      "      #lift# _n _m (#subst# _arg' _p _arg) =" ::
      "      #subst# (#source_lift# _n _m _arg') (_p + _n) (#lift# _n _m _arg)" ::
      []
	
	(** the lemma proof script *)
    method lemma_proof : string list =
      "    #by_induction#; intros;" ::
      "#rewrite_induction_hypothesis#" ::
      "      lift_tac; #crush#." ::
      []
	
	(** extra definition lines to be inserted after the property *)
    method aux_defs : string list =
      []
	
	(** the tactic template : pattern, body and body postpone *)
    method tactic_template : string * (string list * string list) =
      ("#lift# ?n ?m (#subst# ?v ?p ?t)",
       ("          (cutrewrite ->" ::
	"            (#lift# n m (#subst# v p t) =" ::
	"             #subst# (#source_lift# n m v) (p + n) (#lift# n m t))#inH#;" ::
	"            [ cfail | apply (#lemma_name# t v n m p); cfail ])" ::
	[],
	"          (cutrewrite ->" ::
	"            (#lift# n m (#subst# v p t) =" ::
	"             #subst# (#source_lift# n m v) (p + n) (#lift# n m t))#inH#;" ::
	"            [ | apply (#lemma_name# t v n m p); cfail ])" ::
	[]))
  end

class subst_lift_distrib_property _gmod =
  object (self)
    inherit subst_property _gmod as super
	
	(** name of the lemma *)
    method lemma_name_type_name (name : string) : string =
      self#subst_f#fun_name_type_name name ^ "_" ^
      self#lift#fun_name_type_name name ^ "_distrib"
					    
    method subst : (string * string) list =
      ("#source_lift#", self#lift#fun_name_type_name self#get_source#get_name) ::
      super#subst
	
	(** the lemma statement *)
    method lemma_header : string list =
      "  #Lemma# #lemma_name# :" ::
      "    forall (_arg : #target_type#) (_arg' : #source_type#) (_n _m _p : nat)," ::
      "      _p <= _m ->" ::
      "      #lift# _n _m (#subst# _arg' _p _arg) =" ::
      "      #subst# (#source_lift# _n _m _arg') _p (#lift# _n (S _m) _arg)" ::
      []
	
	(** the lemma proof script *)
    method lemma_proof : string list =
      "    #by_induction#; intros;" ::
      "#rewrite_induction_hypothesis#" ::
      "      lift_tac; #crush#." ::
      []
	
	(** extra definition lines to be inserted after the property *)
    method aux_defs : string list =
      []
	
	(** the tactic template : pattern, body and body postpone *)
    method tactic_template : string * (string list * string list) =
      ("#lift# ?n ?m (#subst# ?v ?p ?t)",
       ("          (cutrewrite ->" ::
	"            (#lift# n m (#subst# v p t) =" ::
	"             #subst# (#source_lift# n m v) p (#lift# n (S m) t))#inH#;" ::
	"            [ cfail | apply (#lemma_name# t v n m p); cfail ])" ::
	[],
	[]))
  end

class subst_lift_pack _gmod =
  object
    inherit virtual_property_pack _gmod
	
	(** the title to insert before the definitions *)
    method get_title : string =
      "Interactions between lifting and substitution"
	
	(** the list of properties of this pack *)
    method properties : property list =
      (new subst_lift_simpl_property gmod :> property) ::
(* TODO utile ?      (new subst_lift_simpl_0_property gmod :> property) ::*)
      (new subst_lift_commut_property gmod :> property) ::
      (new subst_lift_distrib_property gmod :> property) ::
      []
	
	(** the tactic name of the tactic associated with this pack,
	 * or None if no tactic is required *)
    method tactic_name : string option = Some "subst_lift_tac"
  end

class subst_subst_distrib_property _gmod =
  object (self)
    inherit subst_property _gmod as super
	
	(** name of the lemma *)
    method lemma_name_type_name (name : string) : string =
      self#subst_f#fun_name_type_name name ^ "_" ^
      self#subst_f#fun_name_type_name name ^ "_distrib"
					       
    method subst : (string * string) list =
      ("#source_lift#", self#lift#fun_name_type_name self#get_source#get_name) ::
      ("#source_subst#", self#subst_f#fun_name_type_name self#get_source#get_name) ::
      super#subst
	
	(** the lemma statement *)
    method lemma_header : string list =
      "  #Lemma# #lemma_name# :" ::
      "    forall (_arg : #target_type#) (_arg' _arg'' : #source_type#) (_n _m : nat)," ::
      "      _m <= _n ->" ::
      "      #subst# _arg' _n (#subst# _arg'' _m _arg) =" ::
      "      #subst# (#source_subst# _arg' _n _arg'') _m (#subst# (#source_lift# 1 _m _arg') (S _n) _arg)" ::
      []
	
	(** the lemma proof script *)
    method lemma_proof : string list =
      "    #by_induction#; intros;" ::
      "#rewrite_induction_hypothesis#" ::
      "      subst_lift_tac; lift_lift_tac; #crush#." ::
      []
	
	(** extra definition lines to be inserted after the property *)
    method aux_defs : string list =
      []
	
	(** the tactic template : pattern, body and body postpone *)
    method tactic_template : string * (string list * string list) =
      ("#subst# ?b ?n (#subst# ?c ?m ?a)",
       ("          (cutrewrite ->" ::
	"            (#subst# b n (#subst# c m a) =" ::
	"             #subst# (#source_subst# b n c) m (#subst# (#source_lift# 1 m b) (S n) a))#inH#;" ::
	"            [ | apply (#lemma_name# a b c n m); cfail ])" ::
	[],
	[]))
  end

class subst_subst_pack _gmod =
  object
    inherit virtual_property_pack _gmod
	
	(** the title to insert before the definitions *)
    method get_title : string =
      "Substitution lemma"
	
	(** the list of properties of this pack *)
    method properties : property list =
      (new subst_subst_distrib_property gmod :> property) ::
      []
	
	(** the tactic name of the tactic associated with this pack,
	 * or None if no tactic is required *)
    method tactic_name : string option = Some "subst_subst_tac"
  end

class subst_chapter _gmod =
  object
    inherit virtual_property_chapter _gmod
	
	(** the title to insert before the definitions *)
    method get_title : string =
      "Substitution main properties"
	
	(** the list of property packs of this chapter *)
    method property_packs : property_pack list =
      (new indep_pack gmod :> property_pack) ::
      (new subst_lift_pack gmod :> property_pack) ::
      (new subst_subst_pack gmod :> property_pack) ::
      []
	
	(** the tactic name of the tactic associated with this chapter,
	 * or None if no tactic is required *)
    method tactic_name : string option = Some "subst_tac"
    method aux_tactics : string list = "subst_index_tac" :: []
  end
