open Xstring
open Xlist
open Xlib

open Interfaces
open Properties_abstract
open Properties_lift_subst
open Definitions_named
open Definitions_lift_subst

class virtual fv_permut_property _gmod =
  object (self)
    inherit lift_property _gmod as super
	
    method fv_permut : fv_permut_fun =
      new Definitions_named.fv_permut_fun_with_source_and_target
	gmod self#get_source self#get_target
	
    method lift : lift_fun =
      new Definitions_lift_subst.lift_fun_with_source_and_target
	gmod self#get_source self#get_target
	
    method subst : (string * string) list =
      ("#fv_permut#", self#fv_permut#fun_name) ::
      ("#lift#", self#lift#fun_name) ::
      []

    method get_index_tac : string = "fv_permut_index_tac; #index_tac#; simpl in *"
  end

(**
 * identity permutation simplification.
 *)
class fv_permut_id_simpl_property _gmod =
  object (self)
    inherit fv_permut_property _gmod as super
	
	(** name of the lemma *)
    method lemma_name_type_name (name : string) : string =
      self#fv_permut#fun_name_type_name name ^ "_id_simpl"
						 
						 (** the lemma statement *)
    method lemma_header : string list =
      "  #Lemma# #lemma_name# :" ::
      "    forall (_k : nat) (_arg : #target_type#) (_m _n : nat)," ::
      "      #fv_permut# _k (List.seq 0 _n) _m _arg = #lift# _k (_m + _n) _arg" ::
      []
	
	(** the lemma proof script *)
    method lemma_proof : string list =
      "    #by_induction#;" ::
      "      intros _m _n;" ::
      "#rewrite_induction_hypothesis#" ::
      "      idtac." ::
      []

    method get_index_tac : string =
      "fv_permut_index_tac; index_tac; #crush#; repeat (progress (rewrite List.seq_nth || rewrite List.seq_length in *))"
	
	(** extra definition lines to be inserted after the property *)
    method aux_defs : string list =
      "  Hint Rewrite #lemma_name# : #database#." ::
      []
	
	(** the tactic template : pattern, body and body postpone *)
    method tactic_template : string * (string list * string list) =
      ("", ([], []))
  end

class fv_permut_id_simpl_pack _gmod =
  object
    inherit virtual_property_pack _gmod
	
	(** the title to insert before the definitions *)
    method get_title : string =
      "Identity free variables permutation simplification"
	
	(** the list of properties of this pack *)
    method properties : property list =
      (new fv_permut_id_simpl_property gmod :> property) ::
      []
  end

(**
 * empty permutation simplification generalization.
 *)
class fv_permut_id_simpl_gen_property _gmod =
  object (self)
    inherit fv_permut_property _gmod as super
	
	(** name of the lemma *)
    method lemma_name_type_name (name : string) : string =
      self#fv_permut#fun_name_type_name name ^ "_id_simpl_gen"
						 
						 (** the lemma statement *)
    method lemma_header : string list =
      "  #Lemma# #lemma_name# :" ::
      "    forall (_k : nat) (_arg : #target_type#) (_l : list nat) (_m _n : nat)," ::
      "      _l = List.seq 0 _n -> #fv_permut# _k _l _m _arg = #lift# _k (_m + List.length _l) _arg" ::
      []
	
	(** the lemma proof script *)
    method lemma_proof : string list =
      "    intros; subst; rewrite List.seq_length; #crush#." ::
      []
	
	(** extra definition lines to be inserted after the property *)
    method aux_defs : string list =
      []
	
	(** the tactic template : pattern, body and body postpone *)
    method tactic_template : string * (string list * string list) =
      ("#fv_permut# ?k ?l ?m ?t",
       ("          (cutrewrite ->" ::
	"            (#fv_permut# k l m t = #lift# k (m + List.length l) t)#inH#;" ::
	"            [ | (apply (#lemma_name# k t l m 0); cfail) ||" ::
	"                (apply (#lemma_name# k t l m 1); cfail) ||" ::
	"                (apply (#lemma_name# k t l m 2); cfail) ||" ::
	"                (apply (#lemma_name# k t l m 3); cfail) ||" ::
	"                (apply (#lemma_name# k t l m 4); cfail) ||" ::
	"                (apply (#lemma_name# k t l m 5); cfail) ])" ::
	[], []))
  end

class fv_permut_id_simpl_gen_pack _gmod =
  object
    inherit virtual_property_pack _gmod
	
	(** the title to insert before the definitions *)
    method get_title : string =
      "Identity free variables permutation simplification generalized"
	
	(** the list of properties of this pack *)
    method properties : property list =
      (new fv_permut_id_simpl_gen_property gmod :> property) ::
      []
	
	(** the tactic name of the tactic associated with this pack,
	 * or None if no tactic is required *)
    method tactic_name : string option = Some "fv_permut_id_simpl_tac"
  end

class fv_permut_chapter _gmod =
  object
    inherit virtual_property_chapter _gmod as super
	
	(** the title to insert before the definitions *)
    method get_title : string =
      "Free variables permutation properties"
	
	(** the list of property packs of this chapter *)
    method property_packs : property_pack list =
      (new fv_permut_id_simpl_pack gmod :> property_pack) ::
      (new fv_permut_id_simpl_gen_pack gmod :> property_pack) ::
      []
	      
	(** the tactic name of the tactic associated with this chapter,
	 * or None if no tactic is required *)
    method tactic_name : string option = Some "fv_permut_tac"
    method aux_tactics : string list = "fv_permut_index_tac" :: []
	
	(** index lemmas *)
    method fun_index_lemmas_and_tactics_to_generate : fun_definition list =
      (new fv_permut_fun gmod) ::
      []
  end
