module Xlist = struct
  
  (**
   * remove all the element equals to an element in a list.
   *
   * @params e l
   * @param e the element to remove.
   * @param l the list.
   *
   * @return the list without any occurence of e.
   *)
  let rec remove e l =
    match l with
      [] -> []
    | e' :: l ->
	let l' = remove e l in
	if e' = e then l' else e' :: l'
				      
				      (**
				       * remove the first element equal to an element in a list.
				       *
				       * @params e l
				       * @param e the element to remove.
				       * @param l the list.
				       *
				       * @return the list without the first occurence of e.
				       *)	
  let rec remove_one e l =
    match l with
      [] -> []
    | e' :: l ->
	if e' = e then l else e' :: (remove_one e l)
				     
				     (**
				      * variant of split for a list of 4-tuples.
				      *
				      * @params l
				      * @param l the list [(a1, b1, c1, d1), ..., (an, bn, cn, dn)].
				      *
				      * @return the 4-tuple of lists
				      * ([a1, ..., an], [b1, ..., bn], [c1, ..., cn], [d1, ..., dn]).
				      *)
  let split_4 l =
    List.fold_right
      (fun (a, b, c, d) -> fun (aa, bb, cc, dd) ->
	(a :: aa, b :: bb, c :: cc, d :: dd))
      l ([], [], [], [])

      (**
       * variant of map that applies the function to the last element of a list.
       *
       * @params f l
       * @param f the function to apply.
       * @param l the list [x1, ..., xn].
       *
       * @return the list [x1, ..., x(n-1), f xn].
       *)
  let rec map_last f l =
    match l with
      [] -> []
    | e :: [] -> (f e) :: []
    | e :: l -> e :: map_last f l

		      (**
		       * variant of map that applies the function to all but last element of a list.
		       *
		       * @params f l
		       * @param f the function to apply.
		       * @param l the list [x1, ..., xn].
		       *
		       * @return the list [f x1, ..., f x(n-1), xn].
		       *)
  let rec map_not_last f l =
    match l with
      [] -> []
    | e :: [] -> e :: []
    | e :: l -> f e :: map_not_last f l
			
			(**
			 * gets all the values associated with a given key in an association list.
			 *
			 * @params key l
			 * @param key the key to look for.
			 * @param l the list [(k1, v1), ..., (kn, vn)].
			 *
			 * @return the list [vj, ..., vk] fo vj ... vk having key in l.
			 *)
  let rec all_assoc key l =
    match l with
      [] -> []
    | (k, v) :: l ->
	let all = all_assoc key l in
	if key = k then v :: all else all
	  
	  (**
	   * removes all associations of a given key in an association list.
	   *
	   * @params key l
	   * @param key the key to remove.
	   * @param l the list [(k1, v1), ..., (kn, vn)].
	   *
	   * @return the list (ki, vi) in the same order such that every ki <> key
	   * for any i.
	   *)
  let rec remove_all_assoc key l =
    match l with
      [] -> []
    | (k, v) :: l ->
	let all = remove_all_assoc key l in
	if key = k then all else (k, v) :: all
					    
					    (**
					     * aggregates every value of the same key in an association list.
					     *
					     * @params f l
					     * @param f the aggregating function that builds an element from 1 or more elements.
					     * @param l the list [(k1, v1), ..., (kn, vn)].
					     *
					     * @return the list (ki, vi) such that every ki are pairwise distinct and
					     * vi = f vj ... vk for vj ... vk having key ki in l.
					     *)
  let rec assoc_aggregate f l =
    match l with
      [] -> []
    | (k, v) :: l ->
	let all_k = v :: (all_assoc k l) in
	let new_l = assoc_aggregate f (remove_all_assoc k l) in
	(k, f all_k) :: new_l
			 
			 (**
			  * calculate the index of a given element in a list.
			  *
			  * @params e l
			  * @param e the element to look for.
			  * @param l the list [x1, ..., xn].
			  *
			  * @return i such that xi = e, raise Not_found if e does not occur in l.
			  *)
  let rec index_of e l =
    match l with
      [] -> raise Not_found
    | x :: l -> if x = e then 0 else 1 + index_of e l
	
	(**
	 * create a list of the given length, filled with elements built by
	 * a function taking the index of the element (starting at 1).
	 *
	 * @params n f
	 * @param n the list length.
	 * @param f the element building function.
	 *
	 * @return the list [f 1, ..., f n].
	 *)
  let rec tabulate n f =
    if n <= 0 then []
    else (f n) :: tabulate (n - 1) f
		   
		   (**
		    * check if the list contains the same elements regardless of the order.
		    *
		    * @params l1 l2
		    * @param l1 the first list.
		    * @param l2 the second list.
		    *
		    * @return true if l1 and l2 contains the same elements (with the same number
		    * of occurence), else false.
		    *)
  let rec multiset_equal l1 l2 =
    match l1 with
      [] -> List.length l2 = 0
    | e :: l1' ->
	if not (List.mem e l2) then false
	else multiset_equal l1' (remove_one e l2)
	    
	    (**
	     * remove doublons in a list.
	     *
	     * @params l
	     * @param l the list.
	     *
	     * @return the list containing all distinct values of l in the same order.
	     *)
  let rec doublon_remove l =
    match l with
      [] -> []
    | e :: l ->
	e :: doublon_remove (remove e l)
	      
	      (**
	       * variant of map, gives to the function the index of the element
	       * starting at 0.
	       *
	       * @params f l
	       * @param f the function to apply.
	       * @param l the list [x0, ..., xn].
	       *
	       * @return [f x0 0, f x1 1, ..., f xn n].
	       *)	
  let rec map_count f l =
    let rec aux n l =
      match l with
	[] -> []
      | e :: l -> (f e n) :: aux (n + 1) l
    in
    aux 0 l
      
      (**
       * variant of map_count with explicit starting number.
       *
       * @params f p l
       * @param f the function to apply.
       * @param p the starting number.
       * @param l the list [x0, ..., xn].
       *
       * @return [f x0 p, f x1 (p + 1), ..., f xn (p + n)].
       *)	
  let rec map_count_start f p l =
    let rec aux n l =
      match l with
	[] -> []
      | e :: l -> (f e n) :: aux (n + 1) l
    in
    aux p l
      
      (**
       * variant of map_count where f can affect the counter.
       *
       * @params f l
       * @param f the function to apply.
       * @param l the list [x0, ..., xn].
       *
       * @return [fst (f x0 0), fst (f x1 (snd (f x0 0) + 1)), ..., fst (f xn (snd f x(n-1) + 1))].
       *)	
  let rec map_count_var f l =
    let rec aux n l =
      match l with
	[] -> []
      | e :: l ->
	  let result = f e n in
	  fst result :: aux (1 + snd result) l
    in
    aux 0 l
      
      (**
       * flattens a list of list, inserting a separator between each inner list
       * and removing empty inner lists.
       *
       * @params sep l
       * @param sep the separator (itself a list).
       * @param l the list of list [l1, l2, ..., ln].
       *
       * @return l1 @ sep @ l2 @ ... @ ln.
       *)
  let rec concat_with lsep ll =
    match ll with
      [] -> []
    | [l] -> l
    | [] :: q -> concat_with lsep q
    | l :: q -> l @ lsep @ concat_with lsep q
			     
			     (**
			      * variant of List.map that takes an additional paraemeter for the head
			      * element and another for every tail elements.
			      *
			      * @params hd_arg tl_arg f l
			      * @param hd_arg the argument to give to the head element of l.
			      * @param tl_arg the argument to give to the tail elements of l
			      * @param f the function to apply.
			      * @param l the list [x1, ..., xn].
			      *
			      * @return [ f hd_arg x1, f tl_arg x2, ... f tl_arg xn ].
			      *)
  let map_with_hd_tl_param hd_arg tl_arg f l =
    match l with
      [] -> []
    | h :: t ->
	(f hd_arg h) :: (List.map (fun x -> f tl_arg x) t)
			 
			 (**
			  * extract a list of value from a list of options.
			  *
			  * @params l
			  * @param l the list [x1, ..., xn].
			  *
			  * @return [xa, ..., xb] where Some xa = xi for some i ; the order is preserved.
			  *)
  let rec option_prune l =
    match l with
      [] -> []
    | Some v :: l -> v :: option_prune l
    | None :: l -> option_prune l
			 
			 (**
			  * builds the ordered combinations of elements of a list.
			  *
			  * @params l
			  * @param l the list [x1, ..., xn].
			  *
			  * @return [ (x1, x2), ..., (x1, xn), (x2, x3), ..., (x2, xn), ..., (xn-1, xn) ].
			  *)
  let build_ordered_combin l =
    let rec aux elem others =
      match others with
	[] -> []
      | e :: others -> (elem, e) :: aux elem others in
    let rec aux2 l =
      match l with
	[] -> []
      | e :: l -> (aux e l) @ aux2 l in
    aux2 l
      
      (**
       * builds the combinations of elements of a list with no repetition.
       *
       * @params l
       * @param l the list [x1, ..., xn].
       *
       * @return [ (x1, x2), ..., (x1, xn), (x2, x1), ..., (x2, xn), ..., (xn-1, x1), ..., xn-1, xn) ] with no (xi, xi).
       *)
  let build_combin all =
    let rec aux elem others =
      match others with
	[] -> []
      | e :: others -> (if elem = e then [] else [(elem, e)]) @ aux elem others in
    let rec aux2 l =
      match l with
	[] -> []
      | e :: l -> (aux e all) @ aux2 l in
    aux2 all
      
      (**
       * concatenate a pair of list.
       *
       * @params lp
       * @param lp the list pair (l1, l2).
       *
       * @return l1 @ l2.
       *)
  let concat_pair lp = (fst lp) @ (snd lp)

end
